package be.iminds.ilabt.jfed.util;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.codec.binary.StringUtils;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.*;
import java.util.zip.DeflaterInputStream;
import java.util.zip.InflaterInputStream;

/**
 * DataConversionUtils
 */
public class DataConversionUtils {
    private static final org.slf4j.Logger LOG = org.slf4j.LoggerFactory.getLogger(DataConversionUtils.class);

    /**
     * note: from command line this can also be done: echo 'DATA' | base64 --decode | openssl zlib -d
     *
     * @param s A base64 encoded string, containing compressed data. Compression is as specified in RFC 1950
     * @return the decoded string, or null if there is an error (!)
     */
    @Nullable
    public static String decompressFromBase64(@Nonnull String s) {
        StringBuilder res = new StringBuilder();
        if (s.length() == 0)
            return s;
        try {
            byte[] b = Base64.decodeBase64(StringUtils.getBytesUtf8(s));

            //chained try-with-resources is only safe for AutoCloseable
            //   (because it calls close() multiple times, and that is only idempotent in AutoCloseable, but not in just Closeable)
            assert AutoCloseable.class.isAssignableFrom(ByteArrayInputStream.class);
            assert AutoCloseable.class.isAssignableFrom(InflaterInputStream.class);
            assert AutoCloseable.class.isAssignableFrom(InputStreamReader.class);
            assert AutoCloseable.class.isAssignableFrom(BufferedReader.class);

            try (ByteArrayInputStream bis = new ByteArrayInputStream(b);
                 InflaterInputStream ifis = new InflaterInputStream(bis);
                 InputStreamReader isr = new InputStreamReader(ifis);
                 BufferedReader br = new BufferedReader(isr)) {

                String line = br.readLine();
                while (line != null) {
                    res.append(line);
                    line = br.readLine();
                }
            }
        } catch (Exception e) {
//            LOG.warn("Could not decompress string of length={}", s.length(), e);
            return null;
        }

        return res.toString();
    }

    /* reverse operation */
    @Nonnull
    public static String compressToBase64(@Nonnull String s) {
        if (s.isEmpty())
            return s;

        assert AutoCloseable.class.isAssignableFrom(ByteArrayInputStream.class);
        assert AutoCloseable.class.isAssignableFrom(DeflaterInputStream.class);
        assert AutoCloseable.class.isAssignableFrom(ByteArrayOutputStream.class);

        byte[] b = StringUtils.getBytesUtf8(s);
        try (ByteArrayInputStream bais = new ByteArrayInputStream(b);
             InputStream is = new DeflaterInputStream(bais);
             ByteArrayOutputStream buffer = new ByteArrayOutputStream()) {

            int nRead;
            byte[] data = new byte[1024];
            while ((nRead = is.read(data, 0, data.length)) != -1) {
                buffer.write(data, 0, nRead);
            }

            return new String(Base64.encodeBase64(buffer.toByteArray()), "UTF-8");
        } catch (Exception e) {
            LOG.warn("Could not compress string of length=" + s.length());
            throw new RuntimeException("Failed to compress", e);
        }
    }

    public static byte[] decodeBase64(String s) {
        return Base64.decodeBase64(StringUtils.getBytesUtf8(s));
    }

    public static String encodeBase64(byte[] b) {
        return Base64.encodeBase64String(b);
    }
}
